@@ -19,7 +19,7 @@ class BipartiteBase(DataFrame):
 
     Arguments:
         *args: arguments for Pandas DataFrame
-        columns_req (list): required columns (only put general column names for joint columns, e.g. put 'fid' instead of 'f1i', 'f2i'; then put the joint columns in reference_dict)
+        columns_req (list): required columns (only put general column names for joint columns, e.g. put 'j' instead of 'j1', 'j2'; then put the joint columns in reference_dict)
         columns_opt (list): optional columns (only put general column names for joint columns, e.g. put 'g' instead of 'g1', 'g2'; then put the joint columns in reference_dict)
         columns_contig (dictionary): columns requiring contiguous ids linked to boolean of whether those ids are contiguous, or None if column(s) not included, e.g. {'i': False, 'j': False, 'g': None} (only put general column names for joint columns)
         reference_dict (dict): clarify which columns are associated with a general column name, e.g. {'i': 'i', 'j': ['j1', 'j2']}
@@ -39,7 +39,7 @@ class BipartiteBase(DataFrame):
         # self.logger.info('initializing BipartiteBase object')
 
         if len(args) > 0 and isinstance(args[0], BipartiteBase): # Note that isinstance works for subclasses
-            self._set_attributes(args[0])
+            self._set_attributes(args[0], include_id_reference_dict)
         else:
             self.columns_req = ['i', 'j', 'y'] + columns_req
             self.columns_opt = ['g', 'm'] + columns_opt
@@ -179,23 +179,24 @@ class BipartiteBase(DataFrame):
         if self.id_reference_dict:
             for id_col, reference_df in self.id_reference_dict.items():
                 if len(reference_df) > 0: # Make sure non-empty
-                    for i, id_subcol in enumerate(to_list(self.reference_dict[id_col])):
+                    for id_subcol in to_list(self.reference_dict[id_col]):
                         try:
-                            frame = frame.merge(reference_df[['original_ids', 'adjusted_ids_' + str(len(reference_df.columns) - 1)]].rename({'original_ids': 'original_' + id_col + '_' + str(i + 1), 'adjusted_ids_' + str(len(reference_df.columns) - 1): id_subcol}, axis=1), how='left', on=id_subcol)
+                            frame = frame.merge(reference_df[['original_ids', 'adjusted_ids_' + str(len(reference_df.columns) - 1)]].rename({'original_ids': 'original_' + id_subcol, 'adjusted_ids_' + str(len(reference_df.columns) - 1): id_subcol}, axis=1), how='left', on=id_subcol)
                         except TypeError: # Int64 error with NaNs
                             frame[id_col] = frame[id_col].astype('Int64')
-                            frame = frame.merge(reference_df[['original_ids', 'adjusted_ids_' + str(len(reference_df.columns) - 1)]].rename({'original_ids': 'original_' + id_col + '_' + str(i + 1), 'adjusted_ids_' + str(len(reference_df.columns) - 1): id_subcol}, axis=1), how='left', on=id_subcol)
+                            frame = frame.merge(reference_df[['original_ids', 'adjusted_ids_' + str(len(reference_df.columns) - 1)]].rename({'original_ids': 'original_' + id_subcol, 'adjusted_ids_' + str(len(reference_df.columns) - 1): id_subcol}, axis=1), how='left', on=id_subcol)
             return frame
         else:
             warnings.warn('id_reference_dict is empty. Either your id columns are already correct, or you did not specify `include_id_reference_dict=True` when initializing your BipartitePandas object')
 
-    def _set_attributes(self, frame, no_dict=False):
+    def _set_attributes(self, frame, no_dict=False, include_id_reference_dict=False):
         '''
         Set class attributes to equal those of another BipartitePandas object.
 
         Arguments:
             frame (BipartitePandas): BipartitePandas object whose attributes to use
             no_dict (bool): if True, only set booleans, no dictionaries
+            include_id_reference_dict (bool): if True, create dictionary of Pandas dataframes linking original id values to contiguous id values
         '''
         # Dictionaries
         if not no_dict:
@@ -205,11 +206,14 @@ class BipartiteBase(DataFrame):
             self.col_dtype_dict = frame.col_dtype_dict.copy()
             self.col_dict = frame.col_dict.copy()
         self.columns_contig = frame.columns_contig.copy() # Required, even if no_dict
-        self.id_reference_dict = {} # Required, even if no_dict
+        self.id_reference_dict = {} # Required, even if no_dict (this goes first, because always need to instantiate the empty dict before it's filled)
         if frame.id_reference_dict:
             # Must do a deep copy
             for id_col, reference_df in frame.id_reference_dict.items():
                 self.id_reference_dict[id_col] = reference_df.copy()
+        elif include_id_reference_dict:
+            # This is if the original dataframe DIDN'T have an id_reference_dict but the new dataframe does
+            self.id_reference_dict = {id_col: pd.DataFrame() for id_col in self.reference_dict.keys()} # Link original id values to contiguous id values
         # Booleans
         self.connected = frame.connected # If False, not connected; if 'connected', all observations are in the largest connected set of firms; if 'biconnected' all observations are in the largest biconnected set of firms; if None, connectedness ignored
         self.correct_cols = frame.correct_cols # If True, column names are correct
@@ -383,7 +387,7 @@ class BipartiteBase(DataFrame):
         Make column of ids contiguous.
 
         Arguments:
-            id_col (str): column to make contiguous ('fid', 'wid', or 'j'). Use general column names for joint columns, e.g. put 'j' instead of 'j1', 'j2'. Only optional columns may be renamed
+            id_col (str): column to make contiguous ('i', 'j', or 'g'). Use general column names for joint columns, e.g. put 'j' instead of 'j1', 'j2'. Only optional columns may be renamed
             copy (bool): if False, avoid copy
 
         Returns:
@@ -684,20 +688,14 @@ class BipartiteBase(DataFrame):
             else:
                 warnings.warn("Trying to create network with 3 or more edges is not possible. Please check df.reference_dict['j']")
             # Find largest connected set of firms
-            if connectedness == 'connected':
-                largest_cc = max(nx.connected_components(G), key=len)
-            elif connectedness == 'biconnected':
-                largest_cc = max(nx.biconnected_components(G), key=len)
-            else:
-                warnings.warn("Invalid connectedness: {}. Valid options are 'connected', 'biconnected', or None.".format(connectedness))
-            # Keep largest connected set of firms
-            if len(to_list(frame.reference_dict['j'])) == 1:
-                outside_cc = frame[(~frame['j'].isin(largest_cc))].shape[0]
-            else:
-                outside_cc = frame[(~frame['j1'].isin(largest_cc)) | (~frame['j2'].isin(largest_cc))].shape[0]
+            connectedness_dict = {
+                'connected': nx.is_connected,
+                'biconnected': nx.is_biconnected
+            }
+            is_connected = connectedness_dict[connectedness](G)
 
-            frame.logger.info('observations outside connected set (should be 0):' + str(outside_cc))
-            if outside_cc > 0:
+            frame.logger.info('data is {}: {}'.format(connectedness, frame.connected))
+            if not is_connected:
                 frame.connected = False
                 success = False
             else:
@@ -888,9 +886,53 @@ class BipartiteBase(DataFrame):
 
         return frame
 
+    def keep_ids(self, id_col, keep_ids, copy=True):
+        '''
+        Only keep ids belonging to a given set of ids.
+
+        Arguments:
+            id_col (str): column of ids to consider ('i', 'j', or 'g')
+            keep_ids (list): ids to keep
+            copy (bool): if False, avoid copy
+
+        Returns:
+            frame (BipartiteBase): BipartiteBase with ids in the given set
+        '''
+        if copy:
+            frame = self.copy()
+        else:
+            frame = self
+
+        for id_subcol in to_list(self.reference_dict[id_col]):
+            frame = frame[frame[id_subcol].isin(keep_ids)]
+
+        return frame
+
+    def drop_ids(self, id_col, drop_ids, copy=True):
+        '''
+        Drop all ids belonging to a given set of ids.
+
+        Arguments:
+            id_col (str): column of ids to consider ('i', 'j', or 'g')
+            drop_ids (list): ids to drop
+            copy (bool): if False, avoid copy
+
+        Returns:
+            frame (BipartiteBase): BipartiteBase with ids in the given set
+        '''
+        if copy:
+            frame = self.copy()
+        else:
+            frame = self
+
+        for id_subcol in to_list(self.reference_dict[id_col]):
+            frame = frame[~(frame[id_subcol].isin(drop_ids))]
+
+        return frame
+
     def min_movers(self, threshold=15, copy=True):
         '''
-        Keep only firms with at least `threshold` many movers.
+        List firms with at least `threshold` many movers.
 
         Arguments:
             threshold (int): minimum number of movers required to keep a firm
@@ -899,12 +941,13 @@ class BipartiteBase(DataFrame):
         Returns:
             valid_firms (list): list of firms with sufficiently many movers
         '''
+        # Generate m column (the function checks if it already exists)
+        self.gen_m()
+
         if copy:
             frame = self.copy()
         else:
             frame = self
-        # Generate m column (the function checks if it already exists)
-        self.gen_m()
 
         frame = frame[frame['m'] == 1] # Keep movers
 
@@ -930,7 +973,7 @@ class BipartiteBase(DataFrame):
 
         return valid_firms
 
-    def attrition_decreasing(self, subsets=np.linspace(0.1, 0.5, 5), threshold=15, user_clean={}, copy=True):
+    def attrition_increasing(self, subsets=np.linspace(0.1, 0.5, 5), threshold=15, user_clean={}, copy=True):
         '''
         First, keep only firms that have at minimum `threshold` many movers. Then take a random subset of subsets[0] percent of remaining movers. Constructively rebuild the data to reach each subsequent value of subsets. Return these subsets as an iterator.
 
@@ -967,18 +1010,18 @@ class BipartiteBase(DataFrame):
         n_wid_drops_1 = int(np.floor((1 - subsets[0]) * len(wids_init))) # Number of wids to drop
         wid_drops_1 = set(np.random.choice(wids_init, size=n_wid_drops_1, replace=False)) # FIXME figure out using RNG seeds
 
-        subset_1 = subset_init[~(subset_init['i'].isin(wid_drops_1))].copy().clean_data(user_clean)
-        subset_1 = subset_1.original_ids(copy=copy)
-        yield subset_1[['i'] + to_list(self.reference_dict['j']) + to_list(self.reference_dict['y']) + to_list(self.reference_dict['t']) + ['m']] # Keep i, j, y, t (drop original_id columns)
+        subset_1 = subset_init[~(subset_init['i'].isin(wid_drops_1))].copy().clean_data()
+        subset_1_orig_ids = subset_1.original_ids(copy=copy)
+        yield subset_1
 
         # Get list of all valid firms
         valid_firms = []
         for i, j_col in enumerate(to_list(self.reference_dict['j'])):
             original_j = 'original_j_' + str(i + 1)
             try: # If working with a subset
-                valid_firms += list(subset_1[original_j].unique())
+                valid_firms += list(subset_1_orig_ids[original_j].unique())
             except KeyError: # If no subset required
-                valid_firms += list(subset_1[j_col].unique())
+                valid_firms += list(subset_1_orig_ids[j_col].unique())
         valid_firms = set(valid_firms)
 
         # Take full list of valid observations
@@ -987,7 +1030,8 @@ class BipartiteBase(DataFrame):
 
         # Determine which wids (for movers) can still be drawn
         all_valid_wids = set(subset_init.loc[subset_init['m'] == 1, 'i'].unique())
-        dropped_wids = all_valid_wids.difference(set(subset_1.loc[subset_1['m'] == 1, 'original_i_1'].unique()))
+        dropped_wids = all_valid_wids.difference(set(subset_1_orig_ids.loc[subset_1_orig_ids['m'] == 1, 'original_i_1'].unique()))
+        del subset_1_orig_ids
         subset_prev = subset_1
 
         for subset_pct in subsets[1:]: # Each step, drop fewer wids
